import { LocalNode, RawControlledAccount, } from "cojson";
import { inspect, } from "../../coValueInterfaces.js";
import { CoMapOf } from "../coMap/coMapOf.js";
import { controlledAccountSym, } from "./account.js";
import * as S from "@effect/schema/Schema";
import { AST, Schema } from "@effect/schema";
import { SharedCoValueConstructor } from "../construction.js";
import { constructorOfSchemaSym } from "../resolution.js";
import { pipeArguments } from "effect/Pipeable";
import { ValueRef } from "../../refs.js";
export function AccountOf(fields) {
    class AccountOfProfileAndRoot extends SharedCoValueConstructor {
        static get ast() {
            return AST.setAnnotation(Schema.instanceOf(this).ast, constructorOfSchemaSym, this);
        }
        static pipe() {
            // eslint-disable-next-line prefer-rest-params
            return pipeArguments(this, arguments);
        }
        constructor(init, options) {
            super();
            if (!("fromRaw" in options)) {
                throw new Error("Can only construct account from raw or with .create()");
            }
            this.isMe =
                options.fromRaw.id == options.fromRaw.core.node.account.id;
            const refs = {
                get profile() {
                    const profileID = options.fromRaw.get("profile");
                    return profileID && new ValueRef(profileID, controlledAccountFromNode(options.fromRaw.core.node), fields.profile);
                },
                get root() {
                    const rootID = options.fromRaw.get("root");
                    return rootID && new ValueRef(rootID, controlledAccountFromNode(options.fromRaw.core.node), fields.root);
                },
            };
            Object.defineProperties(this, {
                id: { value: options.fromRaw.id, enumerable: false },
                _type: { value: "Account", enumerable: false },
                _owner: { value: this, enumerable: false },
                _refs: { value: refs, enumerable: false },
                _raw: { value: options.fromRaw, enumerable: false },
                _loadedAs: {
                    get: () => this.isMe
                        ? this
                        : controlledAccountFromNode(options.fromRaw.core.node),
                    enumerable: false,
                },
                _schema: { value: AccountOfProfileAndRoot, enumerable: false },
            });
            if (this.isMe) {
                this.sessionID =
                    options.fromRaw.core.node.currentSessionID;
            }
        }
        static fromRaw(raw) {
            return new AccountOfProfileAndRoot(undefined, {
                fromRaw: raw,
            });
        }
        static async create(options) {
            const { node } = await LocalNode.withNewlyCreatedAccount({
                ...options,
                migration: options.migration &&
                    (async (rawAccount) => {
                        const account = new AccountOfProfileAndRoot(undefined, {
                            fromRaw: rawAccount,
                        });
                        await options.migration(account);
                    }),
            });
            return new AccountOfProfileAndRoot(undefined, {
                fromRaw: node.account,
            });
        }
        static async become(options) {
            const node = await LocalNode.withLoadedAccount({
                accountID: options.accountID,
                accountSecret: options.accountSecret,
                sessionID: options.sessionID,
                peersToLoadFrom: options.peersToLoadFrom,
                migration: options.migration &&
                    (async (rawAccount) => {
                        const account = new AccountOfProfileAndRoot(undefined, {
                            fromRaw: rawAccount,
                        });
                        await options.migration(account);
                    }),
            });
            return new AccountOfProfileAndRoot(undefined, {
                fromRaw: node.account,
            });
        }
        async acceptInvite(valueID, inviteSecret, valueSchema) {
            if (!this.isMe) {
                throw new Error("Only a controlled account can accept invites");
            }
            await this._raw.acceptInvite(valueID, inviteSecret);
            return valueSchema.load(valueID, { as: this });
        }
        get profile() {
            return this._refs.profile.accessFrom(this);
        }
        get root() {
            return this._refs.root.accessFrom(this);
        }
        myRole() {
            return "admin";
        }
        toJSON() {
            return {
                profile: this.profile?.toJSON(),
                root: this.root?.toJSON(),
            };
        }
        [(Schema.TypeId, inspect)]() {
            return this.toJSON();
        }
        static as() {
            return this;
        }
    }
    AccountOfProfileAndRoot.type = "Account";
    return AccountOfProfileAndRoot;
}
export class BaseProfile extends CoMapOf({
    name: S.string,
}).as() {
}
export class Account extends AccountOf({
    profile: BaseProfile,
    root: S.null,
}).as() {
}
const simpleControlledAccounts = new WeakMap();
export function controlledAccountFromNode(node) {
    if (!(node.account instanceof RawControlledAccount)) {
        throw new Error("Expected a controlled account");
    }
    let simpleAccount;
    if (simpleControlledAccounts.has(node.account)) {
        simpleAccount = simpleControlledAccounts.get(node.account);
    }
    else {
        simpleAccount = Account.fromRaw(node.account);
        simpleControlledAccounts.set(node.account, simpleAccount);
    }
    return simpleAccount;
}
//# sourceMappingURL=accountOf.js.map