import { RawAccount, cojsonInternals, } from "cojson";
import { inspect } from "../../coValueInterfaces.js";
import { AST, Schema } from "@effect/schema";
import { constructorOfSchemaSym, propertyIsCoValueSchema, } from "../resolution.js";
import { ValueRef, makeRefs } from "../../refs.js";
import { Account, controlledAccountFromNode } from "../account/accountOf.js";
import { subscriptionsScopes } from "../../subscriptionScope.js";
import { SharedCoValueConstructor } from "../construction.js";
import { pipeArguments } from "effect/Pipeable";
import { Group } from "../group/groupOf.js";
export function CoListOf(itemSchema) {
    const decodeItem = Schema.decodeSync(itemSchema);
    const encodeItem = Schema.encodeSync(itemSchema);
    class CoListOfItem extends Array {
        static get ast() {
            return AST.setAnnotation(Schema.instanceOf(this).ast, constructorOfSchemaSym, this);
        }
        static pipe() {
            // eslint-disable-next-line prefer-rest-params
            return pipeArguments(this, arguments);
        }
        constructor(init, options) {
            if (typeof init === "number") {
                // this might be called from an intrinsic, like map, trying to create an empty array
                // passing `0` as the only parameter
                // eslint-disable-next-line @typescript-eslint/no-explicit-any
                return new Array(init);
            }
            super();
            const itemsAreCoValues = propertyIsCoValueSchema(itemSchema);
            if (!options) {
                throw new Error("Must provide options");
            }
            let raw;
            if ("fromRaw" in options) {
                raw = options.fromRaw;
            }
            else {
                const rawOwner = options.owner._raw;
                const rawInit = itemsAreCoValues
                    ? init?.map((item) => item.id)
                    : init?.map((item) => encodeItem(item));
                raw = rawOwner.createList(rawInit);
            }
            const refs = itemsAreCoValues
                ? makeRefs((idx) => raw.get(idx), () => Array.from({ length: raw.entries().length }, (_, idx) => idx), controlledAccountFromNode(raw.core.node), (_idx) => itemSchema)
                : [];
            const getEdits = () => new Proxy(this, {
                get: (target, key, receiver) => {
                    if (typeof key === "string" && !isNaN(+key)) {
                        const rawEdit = raw.editAt(Number(key));
                        if (!rawEdit)
                            return undefined;
                        return {
                            get value() {
                                if (itemsAreCoValues) {
                                    return this.ref?.accessFrom(receiver);
                                }
                                else {
                                    return (rawEdit?.value &&
                                        decodeItem(rawEdit?.value));
                                }
                            },
                            get ref() {
                                if (itemsAreCoValues) {
                                    return (rawEdit?.value &&
                                        new ValueRef(rawEdit?.value, receiver._loadedAs, itemSchema));
                                }
                            },
                            get by() {
                                if (cojsonInternals.isAccountID(rawEdit.by)) {
                                    return new ValueRef(rawEdit.by, receiver._loadedAs, Account).accessFrom(receiver);
                                }
                            },
                            madeAt: rawEdit.at,
                            tx: rawEdit.tx,
                        };
                    }
                    else if (key === "length") {
                        return raw.entries().length;
                    }
                    return Reflect.get(target, key, receiver);
                },
            });
            Object.defineProperties(this, {
                id: { value: raw.id, enumerable: false },
                _type: { value: "CoList", enumerable: false },
                _owner: {
                    get: () => raw.group instanceof RawAccount
                        ? Account.fromRaw(raw.group)
                        : Group.fromRaw(raw.group),
                    enumerable: false,
                },
                _refs: { value: refs, enumerable: false },
                _edits: { get: getEdits, enumerable: false },
                _raw: { value: raw, enumerable: false },
                _loadedAs: {
                    get: () => controlledAccountFromNode(raw.core.node),
                    enumerable: false,
                },
                _schema: { value: this.constructor, enumerable: false },
            });
            return new Proxy(this, {
                get(target, key, receiver) {
                    if (typeof key === "string" && !isNaN(+key)) {
                        if (itemsAreCoValues) {
                            return target._refs[Number(key)]?.accessFrom(receiver);
                        }
                        else {
                            return decodeItem(raw.get(Number(key)));
                        }
                    }
                    else if (key === "length") {
                        return raw.entries().length;
                    }
                    const prop = Reflect.get(target, key, receiver);
                    if (typeof prop === "function") {
                        return prop.bind(receiver);
                    }
                    return prop;
                },
                set(target, key, value, receiver) {
                    if (typeof key === "string" && !isNaN(+key)) {
                        if (itemsAreCoValues) {
                            raw.replace(Number(key), value.id);
                            subscriptionsScopes
                                .get(receiver)
                                ?.onRefAccessedOrSet(value.id);
                            return true;
                        }
                        else {
                            raw.replace(Number(key), encodeItem(value));
                            return true;
                        }
                    }
                    else {
                        return Reflect.set(target, key, value);
                    }
                },
                has(target, key) {
                    if (typeof key === "string" && !isNaN(+key)) {
                        return Number(key) < raw.entries().length;
                    }
                    else {
                        return Reflect.has(target, key);
                    }
                },
            });
        }
        static fromRaw(raw) {
            return new CoListOfItem(undefined, { fromRaw: raw });
        }
        push(...items) {
            let rawItems;
            if (propertyIsCoValueSchema(itemSchema)) {
                rawItems = items.map((item) => item.id);
            }
            else {
                rawItems = items.map((item) => encodeItem(item));
            }
            for (const item of rawItems) {
                this._raw.append(item);
            }
            return this._raw.entries().length;
        }
        unshift(...items) {
            let rawItems;
            if (propertyIsCoValueSchema(itemSchema)) {
                rawItems = items.map((item) => item.id);
            }
            else {
                rawItems = items.map((item) => encodeItem(item));
            }
            for (const item of rawItems) {
                this._raw.prepend(item);
            }
            return this._raw.entries().length;
        }
        pop() {
            const last = this[this.length - 1];
            this._raw.delete(this.length - 1);
            return last;
        }
        shift() {
            const first = this[0];
            this._raw.delete(0);
            return first;
        }
        splice(start, deleteCount, ...items) {
            const deleted = this.slice(start, start + deleteCount);
            for (let idxToDelete = start + deleteCount; idxToDelete > start; idxToDelete--) {
                this._raw.delete(idxToDelete);
            }
            let rawItems;
            if (propertyIsCoValueSchema(itemSchema)) {
                rawItems = items.map((item) => item.id);
            }
            else {
                rawItems = items.map((item) => encodeItem(item));
            }
            let appendAfter = start;
            for (const item of rawItems) {
                this._raw.append(item, appendAfter);
                appendAfter++;
            }
            return deleted;
        }
        toJSON() {
            return this.map((item) => typeof item === "object" &&
                "toJSON" in item &&
                typeof item.toJSON === "function"
                ? item.toJSON()
                : item);
        }
        [(Schema.TypeId, inspect)]() {
            return this.toJSON();
        }
        static as() {
            return CoListOfItem;
        }
        subscribe(listener) {
            return SharedCoValueConstructor.prototype.subscribe.call(this, listener);
        }
        subscribeEf() {
            return SharedCoValueConstructor.prototype.subscribeEf.call(this);
        }
    }
    CoListOfItem.type = "CoList";
    CoListOfItem.load = SharedCoValueConstructor.load;
    CoListOfItem.loadEf = SharedCoValueConstructor.loadEf;
    CoListOfItem.subscribe = SharedCoValueConstructor.subscribe;
    CoListOfItem.subscribeEf = SharedCoValueConstructor.subscribeEf;
    return CoListOfItem;
}
//# sourceMappingURL=coListOf.js.map