import { isCoValueSchema, isCoValue, inspect, } from "../../coValueInterfaces.js";
import { ValueRef, makeRefs } from "../../refs.js";
import { RawAccount, cojsonInternals } from "cojson";
import { Account, controlledAccountFromNode } from "../account/accountOf.js";
import { subscriptionsScopes } from "../../subscriptionScope.js";
import { AST, Schema } from "@effect/schema";
import { constructorOfSchemaSym, propertyIsCoValueSchema, } from "../resolution.js";
import { SharedCoValueConstructor } from "../construction.js";
import { propSigToSchema, } from "../../schemaHelpers.js";
import { pipeArguments } from "effect/Pipeable";
import { Group } from "../group/groupOf.js";
export function CoMapOf(fields, indexSignature) {
    class CoMapOfFields extends SharedCoValueConstructor {
        static get ast() {
            return AST.setAnnotation(Schema.instanceOf(this).ast, constructorOfSchemaSym, this);
        }
        static pipe() {
            // eslint-disable-next-line prefer-rest-params
            return pipeArguments(this, arguments);
        }
        constructor(init, options) {
            super();
            let raw;
            if ("fromRaw" in options) {
                raw = options.fromRaw;
            }
            else {
                const rawOwner = options.owner._raw;
                const rawInit = {};
                if (init)
                    for (const key of Object.keys(init)) {
                        const initValue = init[key];
                        const field = fields[key] ||
                            (indexSignature &&
                                Schema.is(indexSignature.key)(key) &&
                                indexSignature.value);
                        if (!field) {
                            throw new Error(`Key ${String(key)} not in schema`);
                        }
                        if (propertyIsCoValueSchema(field)) {
                            // TOOD: check for alignment of actual value with schema
                            if (isCoValue(initValue)) {
                                rawInit[key] = initValue.id;
                            }
                            else {
                                throw new Error(`Expected covalue in ${String(key)} but got ${initValue}`);
                            }
                        }
                        else {
                            const fieldSchema = ("propertySignatureAST" in field
                                ? Schema.make(field.propertySignatureAST.from)
                                : field);
                            rawInit[key] =
                                Schema.encodeSync(fieldSchema)(initValue);
                        }
                    }
                raw = rawOwner.createMap(rawInit);
            }
            for (const key of Object.keys(fields)) {
                if (typeof key !== "string")
                    continue;
                const fieldSchema = fields[key];
                if (propertyIsCoValueSchema(fieldSchema)) {
                    Object.defineProperty(this, key, {
                        get() {
                            return this.getCoValueAtKey(key);
                        },
                        set(value) {
                            this.setCoValueAtKey(key, value);
                        },
                        enumerable: true,
                    });
                }
                else {
                    Object.defineProperty(this, key, {
                        get() {
                            return this.getPrimitiveAtKey(key, fieldSchema);
                        },
                        set(value) {
                            this.setPrimitiveAtKey(key, value, fieldSchema);
                        },
                        enumerable: true,
                    });
                }
            }
            const fieldsThatAreRefs = Object.entries(fields)
                .filter(([_, value]) => isCoValueSchema(value))
                .map(([key]) => key);
            const refs = makeRefs((key) => {
                return raw.get(key);
            }, () => {
                if (indexSignature) {
                    return raw
                        .keys()
                        .filter((key) => fieldsThatAreRefs.includes(key) ||
                        Schema.is(indexSignature.key)(key));
                }
                else {
                    return fieldsThatAreRefs;
                }
            }, controlledAccountFromNode(raw.core.node), (key) => fields[key] ||
                (indexSignature
                    ? indexSignature.value
                    : (() => {
                        throw new Error("Expected " + String(key) + " to be valid");
                    })()));
            const getEdits = () => new Proxy(this, {
                get(target, key, receiver) {
                    const schemaAtKey = fields[key] ||
                        (indexSignature &&
                            Schema.is(indexSignature?.key)(key)
                            ? indexSignature?.value
                            : undefined);
                    if (!schemaAtKey)
                        return Reflect.get(target, key, receiver);
                    const rawEdit = raw.lastEditAt(key);
                    if (!rawEdit)
                        return undefined;
                    return {
                        get value() {
                            if (isCoValueSchema(schemaAtKey)) {
                                return this.ref?.accessFrom(target);
                            }
                            else {
                                return (rawEdit?.value &&
                                    Schema.decodeSync(propSigToSchema(schemaAtKey))(rawEdit?.value));
                            }
                        },
                        get ref() {
                            if (isCoValueSchema(schemaAtKey)) {
                                return (rawEdit?.value &&
                                    new ValueRef(rawEdit?.value, target._loadedAs, schemaAtKey));
                            }
                        },
                        get by() {
                            if (cojsonInternals.isAccountID(rawEdit.by)) {
                                return new ValueRef(rawEdit.by, target._loadedAs, Account).accessFrom(target);
                            }
                        },
                        madeAt: rawEdit.at,
                        tx: rawEdit.tx,
                    };
                },
            });
            Object.defineProperties(this, {
                id: {
                    value: raw.id,
                    enumerable: false,
                },
                _type: { value: "CoMap", enumerable: false },
                _owner: {
                    get: () => raw.group instanceof RawAccount
                        ? Account.fromRaw(raw.group)
                        : Group.fromRaw(raw.group),
                    enumerable: false,
                },
                _refs: { value: refs, enumerable: false },
                _edits: { get: getEdits, enumerable: false },
                _raw: { value: raw, enumerable: false },
                _loadedAs: {
                    get: () => controlledAccountFromNode(raw.core.node),
                    enumerable: false,
                },
                _schema: {
                    value: this.constructor,
                    enumerable: false,
                },
            });
            if (indexSignature) {
                return new Proxy(this, {
                    get(target, key) {
                        if (key in target) {
                            return Reflect.get(target, key);
                        }
                        else {
                            if (Schema.is(indexSignature.key)(key)) {
                                if (propertyIsCoValueSchema(indexSignature.value)) {
                                    return target.getCoValueAtKey(key);
                                }
                                else {
                                    return target.getPrimitiveAtKey(key, indexSignature.value);
                                }
                            }
                            else {
                                throw new Error("No such key " + String(key) + " in CoMap");
                            }
                        }
                    },
                    set(target, key, value) {
                        if (key in target) {
                            return Reflect.set(target, key, value);
                        }
                        else if (Schema.is(indexSignature.key)(key)) {
                            if (propertyIsCoValueSchema(indexSignature.value)) {
                                target.setCoValueAtKey(key, value);
                            }
                            else {
                                target.setPrimitiveAtKey(key, value, indexSignature.value);
                            }
                            return true;
                        }
                        else {
                            throw new Error("No such key " + String(key) + " in CoMap");
                        }
                    },
                    ownKeys(target) {
                        const keys = Reflect.ownKeys(target);
                        for (const key of target._raw.keys()) {
                            if (!keys.includes(key)) {
                                if (Schema.is(indexSignature.key)(key)) {
                                    keys.push(key);
                                }
                            }
                        }
                        return keys;
                    },
                    getOwnPropertyDescriptor(target, key) {
                        if (key in target) {
                            return Reflect.getOwnPropertyDescriptor(target, key);
                        }
                        else if (Schema.is(indexSignature.key)(key)) {
                            return {
                                enumerable: true,
                                configurable: true,
                            };
                        }
                    },
                });
            }
        }
        static fromRaw(raw) {
            return new CoMapOfFields(undefined, { fromRaw: raw });
        }
        getCoValueAtKey(key) {
            return this._refs[key]?.accessFrom(this);
        }
        setCoValueAtKey(key, value) {
            this._raw.set(key, value.id);
            subscriptionsScopes.get(this)?.onRefAccessedOrSet(value.id);
        }
        getPrimitiveAtKey(key, schemaAtKey) {
            return Schema.decodeSync(propSigToSchema(schemaAtKey))(this._raw.get(key));
        }
        setPrimitiveAtKey(key, value, schemaAtKey) {
            this._raw.set(key, Schema.encodeSync(propSigToSchema(schemaAtKey))(value));
        }
        toJSON() {
            return Object.fromEntries(Object.entries(this).flatMap(([key, value]) => typeof value === "object" && "toJSON" in value
                ? [[key, value?.toJSON()]]
                : [[key, value]]));
        }
        [(Schema.TypeId, inspect)]() {
            return this.toJSON();
        }
        static as() {
            return CoMapOfFields;
        }
    }
    CoMapOfFields.type = "CoMap";
    return CoMapOfFields;
}
//# sourceMappingURL=coMapOf.js.map