import { expect, describe, test, beforeEach } from "vitest";
import { webcrypto } from "node:crypto";
import { connectedPeers } from "cojson/src/streamUtils.js";
import { newRandomSessionID } from "cojson/src/coValueCore.js";
import { Effect, Queue } from "effect";
import { Co, S, Account, jazzReady } from "..";
if (!("crypto" in globalThis)) {
    // eslint-disable-next-line @typescript-eslint/no-explicit-any
    globalThis.crypto = webcrypto;
}
beforeEach(async () => {
    await jazzReady;
});
describe("Simple CoMap operations", async () => {
    const me = await Account.create({
        name: "Hermes Puggington",
    });
    class TestMap extends Co.map({
        color: S.string,
        height: S.number,
        birthday: S.Date,
        name: S.optional(S.string),
    }).as() {
        get roughColor() {
            return this.color + "ish";
        }
    }
    const birthday = new Date();
    const map = new TestMap({
        color: "red",
        height: 10,
        birthday: birthday,
    }, { owner: me });
    test("Construction", () => {
        expect(map.color).toEqual("red");
        expect(map.roughColor).toEqual("redish");
        expect(map.height).toEqual(10);
        expect(map.birthday).toEqual(birthday);
        expect(map._raw.get("birthday")).toEqual(birthday.toISOString());
    });
    describe("Mutation", () => {
        test("assignment", () => {
            map.color = "blue";
            expect(map.color).toEqual("blue");
            expect(map._raw.get("color")).toEqual("blue");
            const newBirthday = new Date();
            map.birthday = newBirthday;
            expect(map.birthday).toEqual(newBirthday);
            expect(map._raw.get("birthday")).toEqual(newBirthday.toISOString());
            Object.assign(map, { color: "green", height: 20 });
            expect(map.color).toEqual("green");
            expect(map._raw.get("color")).toEqual("green");
            expect(map.height).toEqual(20);
            expect(map._raw.get("height")).toEqual(20);
            map.name = "Secret name";
            expect(map.name).toEqual("Secret name");
            delete map.name;
            expect(map.name).toEqual(undefined);
        });
    });
});
describe("CoMap resolution", async () => {
    class TwiceNestedMap extends Co.map({
        taste: S.string,
    }).as() {
    }
    class NestedMap extends Co.map({
        name: S.string,
        twiceNested: TwiceNestedMap,
    }).as() {
        get fancyName() {
            return "Sir " + this.name;
        }
    }
    class TestMap extends Co.map({
        color: S.string,
        height: S.number,
        nested: NestedMap,
    }).as() {
        get roughColor() {
            return this.color + "ish";
        }
    }
    const initNodeAndMap = async () => {
        const me = await Account.create({
            name: "Hermes Puggington",
        });
        const map = new TestMap({
            color: "red",
            height: 10,
            nested: new NestedMap({
                name: "nested",
                twiceNested: new TwiceNestedMap({ taste: "sour" }, { owner: me }),
            }, { owner: me }),
        }, { owner: me });
        return { me, map };
    };
    test("Construction", async () => {
        const { map } = await initNodeAndMap();
        // const test: Schema.Schema.To<typeof NestedMap>
        expect(map.color).toEqual("red");
        expect(map.roughColor).toEqual("redish");
        expect(map.height).toEqual(10);
        expect(map.nested?.name).toEqual("nested");
        expect(map.nested?.fancyName).toEqual("Sir nested");
        expect(map.nested?.id).toBeDefined();
        expect(map.nested?.twiceNested?.taste).toEqual("sour");
    });
    test("Loading and availability", async () => {
        const { me, map } = await initNodeAndMap();
        const [initialAsPeer, secondPeer] = connectedPeers("initial", "second", { peer1role: "server", peer2role: "client" });
        me._raw.core.node.syncManager.addPeer(secondPeer);
        const meOnSecondPeer = await Account.become({
            accountID: me.id,
            accountSecret: me._raw.agentSecret,
            peersToLoadFrom: [initialAsPeer],
            sessionID: newRandomSessionID(me.id),
        });
        const loadedMap = await TestMap.load(map.id, { as: meOnSecondPeer });
        expect(loadedMap?.color).toEqual("red");
        expect(loadedMap?.height).toEqual(10);
        expect(loadedMap?.nested).toEqual(undefined);
        expect(loadedMap?._refs.nested?.id).toEqual(map.nested?.id);
        expect(loadedMap?._refs.nested?.value).toEqual(undefined);
        const loadedNestedMap = await NestedMap.load(map.nested.id, {
            as: meOnSecondPeer,
        });
        expect(loadedMap?.nested?.name).toEqual("nested");
        expect(loadedMap?.nested.fancyName).toEqual("Sir nested");
        expect(loadedMap?._refs.nested?.value).toEqual(loadedNestedMap);
        expect(loadedMap?.nested?.twiceNested?.taste).toEqual(undefined);
        const loadedTwiceNestedMap = await TwiceNestedMap.load(map.nested.twiceNested.id, { as: meOnSecondPeer });
        expect(loadedMap?.nested?.twiceNested?.taste).toEqual("sour");
        expect(loadedMap?.nested?._refs.twiceNested?.value).toEqual(loadedTwiceNestedMap);
        const otherNestedMap = new NestedMap({
            name: "otherNested",
            twiceNested: new TwiceNestedMap({ taste: "sweet" }, { owner: meOnSecondPeer }),
        }, { owner: meOnSecondPeer });
        loadedMap.nested = otherNestedMap;
        expect(loadedMap?.nested?.name).toEqual("otherNested");
        expect(loadedMap?._refs.nested?.id).toEqual(otherNestedMap.id);
        expect(loadedMap?._refs.nested?.value).toEqual(otherNestedMap);
        expect(loadedMap?.nested?.twiceNested?.taste).toEqual("sweet");
        expect(loadedMap?.nested?._refs.twiceNested?.value).toBeDefined();
    });
    test("Subscription & auto-resolution", async () => {
        const { me, map } = await initNodeAndMap();
        const [initialAsPeer, secondAsPeer] = connectedPeers("initial", "second", { peer1role: "server", peer2role: "client" });
        me._raw.core.node.syncManager.addPeer(secondAsPeer);
        const meOnSecondPeer = await Account.become({
            accountID: me.id,
            accountSecret: me._raw.agentSecret,
            peersToLoadFrom: [initialAsPeer],
            sessionID: newRandomSessionID(me.id),
        });
        await Effect.runPromise(Effect.gen(function* ($) {
            const queue = yield* $(Queue.unbounded());
            TestMap.subscribe(map.id, { as: meOnSecondPeer }, (subscribedMap) => {
                console.log("subscribedMap.nested?.twiceNested?.taste", subscribedMap.nested?.twiceNested?.taste);
                Effect.runPromise(Queue.offer(queue, subscribedMap));
            });
            const update1 = yield* $(Queue.take(queue));
            expect(update1.nested).toEqual(undefined);
            const update2 = yield* $(Queue.take(queue));
            expect(update2.nested?.name).toEqual("nested");
            map.nested.name = "nestedUpdated";
            const _ = yield* $(Queue.take(queue));
            const update3 = yield* $(Queue.take(queue));
            expect(update3.nested?.name).toEqual("nestedUpdated");
            const oldTwiceNested = update3.nested.twiceNested;
            expect(oldTwiceNested?.taste).toEqual("sour");
            // When assigning a new nested value, we get an update
            const newTwiceNested = new TwiceNestedMap({
                taste: "sweet",
            }, { owner: meOnSecondPeer });
            const newNested = new NestedMap({
                name: "newNested",
                twiceNested: newTwiceNested,
            }, { owner: meOnSecondPeer });
            update3.nested = newNested;
            yield* $(Queue.take(queue));
            // const update4 = yield* $(Queue.take(queue));
            const update4b = yield* $(Queue.take(queue));
            expect(update4b.nested?.name).toEqual("newNested");
            expect(update4b.nested?.twiceNested?.taste).toEqual("sweet");
            // we get updates when the new nested value changes
            newTwiceNested.taste = "salty";
            const update5 = yield* $(Queue.take(queue));
            expect(update5.nested?.twiceNested?.taste).toEqual("salty");
            newTwiceNested.taste = "umami";
            const update6 = yield* $(Queue.take(queue));
            expect(update6.nested?.twiceNested?.taste).toEqual("umami");
        }));
    });
    class TestMapWithOptionalRef extends Co.map({
        color: S.string,
        nested: S.optional(NestedMap),
    }).as() {
    }
    test("Construction with optional", async () => {
        const me = await Account.create({
            name: "Hermes Puggington",
        });
        const mapWithout = new TestMapWithOptionalRef({
            color: "red",
        }, { owner: me });
        expect(mapWithout.color).toEqual("red");
        expect(mapWithout.nested).toEqual(undefined);
        const mapWith = new TestMapWithOptionalRef({
            color: "red",
            nested: new NestedMap({
                name: "wow!",
                twiceNested: new TwiceNestedMap({ taste: "sour" }, { owner: me }),
            }, { owner: me }),
        }, { owner: me });
        expect(mapWith.color).toEqual("red");
        expect(mapWith.nested?.name).toEqual("wow!");
        expect(mapWith.nested?.fancyName).toEqual("Sir wow!");
        expect(mapWith.nested?._raw).toBeDefined();
    });
    class TestRecord extends Co.map({ color: S.string }, { key: S.string, value: S.string }).as() {
    }
    test("Construction with index signature", async () => {
        const me = await Account.create({
            name: "Hermes Puggington",
        });
        const record = new TestRecord({
            color: "red",
            other: "wild",
        }, { owner: me });
        expect(record.color).toEqual("red");
        expect(record._raw.get("color")).toEqual("red");
        expect(record.other).toEqual("wild");
        expect(record._raw.get("other")).toEqual("wild");
        expect(Object.keys(record)).toEqual(["color", "other"]);
    });
});
//# sourceMappingURL=coMap.test.js.map